from typing import Callable, Dict

import numpy as np
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from trading_models.utils import plot_general


def round_dx(x, dx):
    return round(round(x / dx) * dx, 10)


def pymoo_minimize(func: Callable, conf: Dict, algo=GA()):
    xl = [v[0] for v in conf.values()]
    xu = [v[1] for v in conf.values()]
    dx = [v[2] for v in conf.values()]
    best = np.inf

    class Prob(ElementwiseProblem):
        def __init__(s):
            super().__init__(n_var=len(xl), n_obj=1, xl=xl, xu=xu)

        def _evaluate(s, X: np.ndarray, out, *args, **kwargs):
            X = [round_dx(xi, dxi) for xi, dxi in zip(X, dx)]
            P = dict(zip(conf.keys(), X))
            loss = func(P)
            nonlocal best
            if loss < best:
                best = loss
                print(f"best: {best} {P}")
            out["F"] = loss

    minimize(Prob(), algo, seed=42)


def plot_trades(trades):
    trades = np.array(trades)
    plots = {
        f"worth ({len(trades)} trades)": trades[:, -1],
        "position_hist": trades[:, 2],
        "duration_hist": trades[:, -3],
        "profit_hist": trades[:, -2],
    }
    plot_general(plots, "simulate")



import numba
import numpy as np


@numba.njit
def simulate(
    sim_data: np.ndarray,
    timeout: int,
    take_profit: float,
    stop_loss: float,
    fee: float,
    init_cash=10e3,
    min_cash=10,
    alloc_ratio=0.01,
    use_ratio=0.2,
    min_pos=0.5,
):
    max_open = int(use_ratio / alloc_ratio)
    worth = cash_left = init_cash
    open_trades = {}
    done_trades = []

    TIME, SYMBOL, FIELD = sim_data.shape
    sim_data[:, :, 2] = np.round(sim_data[:, :, 2], 2)

    for t in range(TIME):

        for id, trade in list(open_trades.items()):
            t1, s1, pos1, price1, cash1 = trade[:5]
            time2, price2, _ = sim_data[t, int(s1)]
            dt = t - t1
            pr = np.sign(pos1) * (price2 / price1 - 1) - fee
            if dt >= timeout or pr >= take_profit or pr <= stop_loss:
                pr = min(pr, take_profit)
                cash_left += cash1 * (1 + pr)
                worth += cash1 * pr
                del open_trades[id]
                trade[-3:] = dt, pr, worth
                done_trades.append(trade)

        for s in range(SYMBOL):
            time, price, pos = sim_data[t, s]
            if abs(pos) >= min_pos and len(open_trades) < max_open:
                cash = int(min(cash_left, worth * alloc_ratio * abs(pos)))
                id = int((s + 1) * np.sign(pos))
                if cash >= min_cash and id not in open_trades:
                    cash_left -= cash
                    open_trades[id] = np.array([t, s, pos, price, cash, 0.0, 0.0, 0.0])

    return done_trades



from typing import Dict

import numpy as np
from crypto_data_downloader.utils import load_pkl
from trading_models.utils import plot_general

from fast_trading_simulator.simulate import simulate

data: Dict = load_pkl("futures_sim_data_2025-07-01_2025-08-01.pkl", gz=True)
sim_data: Dict[str, Dict[str, np.ndarray]] = data["sim_data"]
# for sym, x in sim_data.items():
#     x["position"] = custom_strategy(x["close"])
symbols = list(sim_data.keys())
fields = list(sim_data["BTCUSDT"].keys())
arr = np.array([list(x.values()) for x in sim_data.values()])
arr = arr.transpose((2, 0, 1))

print(f"data keys: {list(data.keys())}")
print(f"arr.shape: {arr.shape} (time, symbols, fields)")
print(f"{len(symbols)} symbols: {symbols[:3]}...")
print(f"{len(fields)} fields: {fields}")
"""
timeout: int, number of time steps
take_profit: float, e.g. 0.01 (1%)
stop_loss: float, e.g. -0.3 (-30%)
fee: float, buy+sell total, e.g. 7e-4 (0.07%)
"""

data["sim_data"] = arr
data.pop("timeout_min", None)
trades = np.array(simulate(**data, init_cash=10e3, alloc_ratio=0.005))
plots = {
    f"worth ({len(trades)} trades)": trades[:, -1],
    "position_hist": trades[:, 2],
    "duration_hist": trades[:, -3],
    "profit_hist": trades[:, -2],
}
plot_general(plots, "simulate")



import numpy as np
import talib as ta
from crypto_data_downloader.utils import load_pkl
from trading_models.utils import D_TYPE

from fast_trading_simulator.simulate import simulate
from fast_trading_simulator.utils import plot_trades, pymoo_minimize

path = "futures_data_2025-08-01_2025-11-20.pkl"
data: D_TYPE = load_pkl(path, gz=True)
T = len(data["BTCUSDT"])
cache = {}


def loss_func(P, plot=False):
    sim_data = []
    for sym, arr in data.items():
        if len(arr) == T:
            time, close = arr.T
            period = int(P["period"])
            id = f"{sym}_{period}"
            x = cache.get(id) or close / ta.KAMA(close, period) - 1
            pos = np.zeros(T)
            pos[x > P["high_bar"]] = -1
            pos[x < P["low_bar"]] = 1
            sim_data.append([time, close, pos])

    sim_data = np.array(sim_data).transpose(2, 0, 1)
    trades = simulate(
        sim_data,
        timeout=P["timeout"],
        take_profit=P["take_profit"],
        stop_loss=P["stop_loss"],
        fee=1e-3,
        alloc_ratio=P["alloc_ratio"],
        use_ratio=P["use_ratio"],
    )
    if plot:
        plot_trades(trades)
    gain = trades[-1][-1] / trades[0][-1]
    return -gain


conf = {
    "period": [20, 100, 1],
    "high_bar": [0.01, 0.1, 0.001],
    "low_bar": [-0.1, -0.01, 0.001],
    "timeout": [10, 100, 1],
    "take_profit": [0.01, 0.1, 0.001],
    "stop_loss": [-1, -0.2, 0.001],
    "alloc_ratio": [0.01, 0.1, 0.001],
    "use_ratio": [0.1, 1, 0.001],
}
P0 = {
    "period": 82,
    "high_bar": 0.092,
    "low_bar": -0.017,
    "timeout": 87,
    "take_profit": 0.096,
    "stop_loss": -0.607,
    "alloc_ratio": 0.1,
    "use_ratio": 0.214,
}
if 0:
    pymoo_minimize(loss_func, conf)
else:
    loss_func(P0, plot=True)



from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="fast-trading-simulator",
    version="0.0.4",
    author="Ricky Ding",
    author_email="e0134117@u.nus.edu",
    description="Numba accelerated minimalist trading simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SerenaTradingResearch/fast-trading-simulator",
    packages=find_packages(),
    install_requires=open("requirements.txt").read().splitlines(),
    python_requires=">=3.8",
    license="MIT",
    keywords=["trading", "simulator", "numba", "quant", "minimalist"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Financial",
    ],
)
